/*
    *
    * This file is a part of CorePDF.
    * A PDF viewer for C Suite.
    * Copyright 2019 CuboCore Group
    *

    *
    * The code contained in this file was originally taken from QtPdfWidgets.
    * Various changes have been made. All bugs are to be reported to CorePDF.
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, vsit http://www.gnu.org/licenses/.
    *
*/

#include "qpdfview.h"
#include "qpdfview_p.h"

#include <QPdfPageRenderer>
#include <QPdfDocument>
#include <QPdfPageNavigation>
#include <QPdfSearchModel>
#include <QPdfSearchResult>

#include <QGuiApplication>
#include <QScreen>
#include <QScrollBar>
#include <QScroller>
#include <QMessageBox>

QT_BEGIN_NAMESPACE

QPdfView::QPdfView( QWidget *parent )
    : QAbstractScrollArea( *new QPdfViewPrivate(), parent )
{
    Q_D( QPdfView );

    d->init();

    /** Piggy-back the matchesFound, and searchComplete signals from the search thread */
    connect( d->m_searchThread, &PdfSearchThread::matchesFound, this, &QPdfView::matchesFound );
    connect( d->m_searchThread, &PdfSearchThread::searchComplete, this, &QPdfView::searchComplete );

    connect(
        d->m_searchThread, &PdfSearchThread::resultsReady, [d, this]( int page ) {
            d->searchRects[ page ] = d->m_searchThread->results( page );
            viewport()->update();
        }
    );

    connect(
        d->m_searchThread, &PdfSearchThread::resultsReady, [d, this]( int page ) {
            d->searchRects[ page ] = d->m_searchThread->results( page );
            viewport()->update();
        }
    );

    connect(
        d->m_pageNavigation, &QPdfPageNavigation::currentPageChanged, this, [d]( int page ){
             d->currentPageChanged( page );
             d->m_searchThread->searchPage( page );
         }
     );

    connect(
        d->m_pageRenderer, &QPdfPageRenderer::pageRendered,
            this, [d]( int pageNumber, QSize imageSize, const QImage &image, QPdfDocumentRenderOptions, quint64 requestId ){
                d->pageRendered( pageNumber, imageSize, image, requestId );
            }
    );

    verticalScrollBar()->setSingleStep( 20 );
    horizontalScrollBar()->setSingleStep( 20 );

    QScroller::grabGesture( this );

    setFrameStyle( QFrame::NoFrame );

    d->calculateViewport();
}

/*!
  \internal
*/
QPdfView::QPdfView( QPdfViewPrivate &dd, QWidget *parent )
    : QAbstractScrollArea( dd, parent )
{
}

QPdfView::~QPdfView()
{
}

void QPdfView::setDocument( QPdfDocument *document )
{
    Q_D( QPdfView );

    if ( d->m_document == document )
        return;

    if ( d->m_document )
        disconnect( d->m_documentStatusChangedConnection );

    d->m_document = document;
    emit documentChanged( d->m_document );

    if ( d->m_document )
        d->m_documentStatusChangedConnection = connect( d->m_document, &QPdfDocument::statusChanged, this, [d](){ d->documentStatusChanged(); } );

    d->m_pageNavigation->setDocument( d->m_document );
    d->m_pageRenderer->setDocument( d->m_document );
    d->m_searchThread->setDocument( document );

    d->documentStatusChanged();
}

QPdfDocument *QPdfView::document() const
{
    Q_D( const QPdfView );

    return d->m_document;
}

QPdfPageNavigation *QPdfView::pageNavigation() const
{
    Q_D( const QPdfView );

    return d->m_pageNavigation;
}

bool QPdfView::isLayoutContinuous() const
{
    Q_D( const QPdfView );

    return d->m_continuous;
}

void QPdfView::setLayoutContinuous( bool yes )
{
    Q_D( QPdfView );

    if ( d->m_continuous == yes )
        return;

    d->m_continuous = yes;
    d->invalidateDocumentLayout();

    emit layoutContinuityChanged( yes );
}

QPdfView::PageLayout QPdfView::pageLayout() const
{
    Q_D( const QPdfView );

    return d->m_pageLayout;
}

void QPdfView::setPageLayout( PageLayout layout )
{
    Q_D( QPdfView );

    if ( d->m_pageLayout == layout )
        return;

    d->m_pageLayout = layout;
    d->invalidateDocumentLayout();

    emit pageLayoutChanged( d->m_pageLayout );
}

QPdfView::ZoomMode QPdfView::zoomMode() const
{
    Q_D( const QPdfView );

    return d->m_zoomMode;
}

void QPdfView::setZoomMode( ZoomMode mode )
{
    Q_D( QPdfView );

    if ( d->m_zoomMode == mode )
        return;

    d->m_zoomMode = mode;
    d->invalidateDocumentLayout();

    emit zoomModeChanged( d->m_zoomMode );
}

qreal QPdfView::zoomFactor() const
{
    Q_D( const QPdfView );

    return d->m_zoomFactor;
}

void QPdfView::setZoomFactor( qreal factor )
{
    Q_D( QPdfView );

    if ( d->m_zoomFactor == factor )
        return;

    d->m_zoomFactor = factor;

    if ( d->m_zoomMode == QPdfView::CustomZoom ) {
        d->invalidateDocumentLayout();
        emit zoomFactorChanged( d->m_zoomFactor );
    }
}

int QPdfView::pageSpacing() const
{
    Q_D( const QPdfView );

    return d->m_pageSpacing;
}

void QPdfView::setPageSpacing( int spacing )
{
    Q_D( QPdfView );

    if ( d->m_pageSpacing == spacing )
        return;

    d->m_pageSpacing = spacing;
    d->invalidateDocumentLayout();

    emit pageSpacingChanged( d->m_pageSpacing );
}

QMargins QPdfView::documentMargins() const
{
    Q_D( const QPdfView );

    return d->m_documentMargins;
}

void QPdfView::setDocumentMargins( QMargins margins )
{
    Q_D( QPdfView );

    if ( d->m_documentMargins == margins )
        return;

    d->m_documentMargins = margins;
    d->invalidateDocumentLayout();

    emit documentMarginsChanged( d->m_documentMargins );
}

void QPdfView::searchText( QString str )
{
    Q_D( QPdfView );

    d->searchRects.clear();

    d->m_searchThread->setSearchString( str );
    d->m_searchThread->searchPage( d->m_pageNavigation->currentPage() );
}

QPdf::Rotation QPdfView::rotation() const
{
    Q_D( const QPdfView );

    return d->m_renderOpts.rotation();
}

void QPdfView::setRotation( QPdf::Rotation rot )
{
    Q_D( QPdfView );

    d->m_renderOpts.setRotation( rot );
    d->invalidateDocumentLayout();

    emit rotationChanged( rot );
}

void QPdfView::paintEvent(QPaintEvent *event)
{
    Q_D(QPdfView);

    QPainter *painter = new QPainter( viewport() );
    painter->fillRect( event->rect(), palette().brush( QPalette::Dark ) );
    painter->translate( -d->m_viewport.x(), -d->m_viewport.y() );

    for( int i = 0; i < d->m_document->pageCount(); i++ ) {

        if ( not d->m_documentLayout.pageGeometries.contains( i ) )
            continue;

        const QRect pgRect = d->m_documentLayout.pageGeometries[ i ];
        if ( pgRect.isNull() )
            continue;

        if ( pgRect.intersects( d->m_viewport ) ) {
            painter->fillRect( pgRect, Qt::white );

            if ( d->m_pageCache.contains( i ) ) {
                /** Get the image */
                QImage page = d->m_pageCache[ i ];

                /** If the image is good, draw it and go to next page */
                if ( not page.isNull() ) {
                    d->paintSearchRects( painter, i );
                    painter->drawImage( pgRect.topLeft(), page );

                    continue;
                }
            }

            /** Page wasn't there or image was not good. */
            d->m_pageRenderer->requestPage( i, pgRect.size(), d->m_renderOpts );
        }
    }

    painter->end();
}

void QPdfView::resizeEvent( QResizeEvent *event )
{
    Q_D( QPdfView );

    QAbstractScrollArea::resizeEvent( event );

    d->updateScrollBars();
    d->calculateViewport();
}

void QPdfView::scrollContentsBy( int dx, int dy )
{
    Q_D( QPdfView );

    QAbstractScrollArea::scrollContentsBy( dx, dy );

    d->calculateViewport();
}

void QPdfView::keyPressEvent( QKeyEvent *kEvent )
{

    Q_D( QPdfView );

    switch( kEvent->key() ) {
        case Qt::Key_Right: {
            /* Go to next page */
            d->m_pageNavigation->setCurrentPage( d->m_pageNavigation->currentPage() + 1 );
            break;
        }

        case Qt::Key_Left: {
            /* Go to previous page */
            d->m_pageNavigation->setCurrentPage( d->m_pageNavigation->currentPage() - 1 );
            break;
        }

        case Qt::Key_Down: {
            /** Single row layout. */
            if ( not d->m_continuous ) {
                /** Scroll down past the end */
                if ( verticalScrollBar()->value() == verticalScrollBar()->maximum() ) {
                    /** Already at the bottom of last page */
                    if ( d->m_pageNavigation->currentPage() == d->m_document->pageCount() - 1 )
                        break;

                    d->m_pageNavigation->setCurrentPage( d->m_pageNavigation->currentPage() + 1 );
                    break;
                }
            }

            QAbstractScrollArea::keyPressEvent( kEvent );
            break;
        }

        case Qt::Key_Up: {
            /** Single row layout. */
            if ( not d->m_continuous ) {

                /** Scroll down past the end */
                if ( verticalScrollBar()->value() == verticalScrollBar()->minimum() ) {

                    /** Already at the top of first page */
                    if ( d->m_pageNavigation->currentPage() == 0 )
                        break;

                    d->m_pageNavigation->setCurrentPage( d->m_pageNavigation->currentPage() - 1 );
                    verticalScrollBar()->setValue( verticalScrollBar()->maximum() );
                    break;
                }
            }

            QAbstractScrollArea::keyPressEvent( kEvent );
            break;
        }

        case Qt::Key_Space: {
            /* Move by approximately 90% of the viewport height */
            verticalScrollBar()->setValue( verticalScrollBar()->value() + viewport()->height() * 0.9 );
            break;
        }

        case Qt::Key_Home: {
            /* Go to first page */
            verticalScrollBar()->setValue( 0 );
            break;
        }

        case Qt::Key_End: {
            /* Go to end of the document */
            verticalScrollBar()->setValue( verticalScrollBar()->maximum() );
            break;
        }

        case Qt::Key_Plus: {
            /* Zoom In */
            setZoomFactor( d->m_zoomFactor * 1.10 );
            break;
        }

        case Qt::Key_Minus: {
            /* Zoom Out */
            setZoomFactor( d->m_zoomFactor / 1.10 );
            break;
        }

        default: {
            /* Default action */
            QAbstractScrollArea::keyPressEvent( kEvent );
            break;
        }
    }
};

void QPdfView::wheelEvent( QWheelEvent *wEvent )
{

    Q_D( QPdfView );

    if ( wEvent->buttons() & Qt::RightButton || (wEvent->modifiers() & Qt::ControlModifier)) {
        QPoint numDegrees = wEvent->angleDelta() / 8;

        int steps = numDegrees.y() / 15;

        QList<qreal> zooms = { 0.1, 0.2, 0.5, 0.75, 1.0, 1.5, 2.0, 4.0 };

        /** - value => scroll up -> increase zoom */
        /** - value => scroll down -> decrease zoom */

        if(d->m_zoomMode == QPdfView::FitToWidth){
            setZoomMode( QPdfView::CustomZoom );
            setZoomFactor(1.0);
        } else if(d->m_zoomMode == QPdfView::FitInView){
            setZoomMode( QPdfView::CustomZoom );
            setZoomFactor(0.5);
        } else{
            setZoomMode( QPdfView::CustomZoom );
        }

        if ( steps < 0 ) {
            if( d->m_zoomFactor >= 0.2 ){
                setZoomFactor( zooms.at( zooms.indexOf( d->m_zoomFactor ) - 1 ) );}
        } else {
            if( d->m_zoomFactor < 4.0 ){
                setZoomFactor( zooms.at( zooms.indexOf( d->m_zoomFactor ) + 1 ) );}
        }

        return;
    }

    /** Single row layout. */
    if ( not d->m_continuous ) {
        QPoint numDegrees = wEvent->angleDelta() / 8;
        int steps = numDegrees.y() / 15;

        /** Scroll down past the end */
        if ( ( verticalScrollBar()->value() == verticalScrollBar()->maximum() ) and ( steps < 0 ) ) {
            d->m_pageNavigation->setCurrentPage( d->m_pageNavigation->currentPage() + 1 );
            return;
        }

        /** Scroll up before the beginning */
        if ( ( verticalScrollBar()->value() == verticalScrollBar()->minimum() ) and ( steps > 0 ) ) {
            d->m_pageNavigation->setCurrentPage( d->m_pageNavigation->currentPage() - 1 );
            return;
        }
    }

    QAbstractScrollArea::wheelEvent( wEvent );
};

QT_END_NAMESPACE
