/*
    *
    * This file is a part of CorePDF.
    * A PDF viewer for C Suite.
    * Copyright 2019 CuboCore Group
    *

    *
    * The code contained in this file was originally taken from QtPdfWidgets.
    * Various changes have been made. All bugs are to be reported to CorePDF.
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, vsit http://www.gnu.org/licenses/.
    *
*/

#ifndef QPDFVIEW_H
#define QPDFVIEW_H

#include <QPdfDocumentRenderOptions>
#include <QtWidgets/qabstractscrollarea.h>

QT_BEGIN_NAMESPACE

class QPdfDocument;
class QPdfPageNavigation;
class QPdfViewPrivate;

class QPdfView : public QAbstractScrollArea
{
    Q_OBJECT;

    Q_PROPERTY(QPdfDocument* document READ document WRITE setDocument NOTIFY documentChanged);

    Q_PROPERTY(PageLayout pageLayout READ pageLayout WRITE setPageLayout NOTIFY pageLayoutChanged);
    Q_PROPERTY(ZoomMode zoomMode READ zoomMode WRITE setZoomMode NOTIFY zoomModeChanged);
    Q_PROPERTY(qreal zoomFactor READ zoomFactor WRITE setZoomFactor NOTIFY zoomFactorChanged);

    Q_PROPERTY(int pageSpacing READ pageSpacing WRITE setPageSpacing NOTIFY pageSpacingChanged);
    Q_PROPERTY(QMargins documentMargins READ documentMargins WRITE setDocumentMargins NOTIFY documentMarginsChanged);

public:

    enum PageLayout
    {
        SinglePage,
        FacingPages,
        BookView,
        OverView
    };
    Q_ENUM(PageLayout)

    enum ZoomMode
    {
        FitToWidth,
        FitInView,
        CustomZoom
    };
    Q_ENUM(ZoomMode)

    explicit QPdfView(QWidget *parent = nullptr);
    ~QPdfView();

    void setDocument(QPdfDocument *document);
    QPdfDocument *document() const;

    QPdfPageNavigation *pageNavigation() const;

    bool isLayoutContinuous() const;
    ZoomMode zoomMode() const;
    qreal zoomFactor() const;
    PageLayout pageLayout() const;

    int pageSpacing() const;
    void setPageSpacing(int spacing);

    QMargins documentMargins() const;
    void setDocumentMargins(QMargins margins);

    void searchText( QString str );

    QPdf::Rotation rotation() const;
    void setRotation( QPdf::Rotation );

public Q_SLOTS:
    void setLayoutContinuous(bool);
    void setZoomMode(ZoomMode mode);
    void setZoomFactor(qreal factor);
    void setPageLayout(PageLayout pgLyt);

Q_SIGNALS:
    void documentChanged(QPdfDocument *document);
    void layoutContinuityChanged(bool continuous);
    void pageLayoutChanged(PageLayout pageLayout);
    void rotationChanged(QPdf::Rotation rotation);
    void zoomModeChanged(ZoomMode zoomMode);
    void zoomFactorChanged(qreal zoomFactor);
    void pageSpacingChanged(int pageSpacing);
    void documentMarginsChanged(QMargins documentMargins);

    void documentLoadingFailed();

    /** Search signals */
    void matchesFound( int numMatches );
    void searchComplete( int numMatches );

protected:
    explicit QPdfView(QPdfViewPrivate &, QWidget *);

    void paintEvent(QPaintEvent *event) override;
    void resizeEvent(QResizeEvent *event) override;
    void scrollContentsBy( int dx, int dy ) override;

    void keyPressEvent( QKeyEvent *kEvent );

    void wheelEvent( QWheelEvent *wEvent );

private:
    Q_DECLARE_PRIVATE(QPdfView)
};

QT_END_NAMESPACE

#endif // QPDFVIEW_H
