/*
    *
    * This file is a part of CorePDF.
    * A PDF viewer for C Suite.
    * Copyright 2019 CuboCore Group
    *

    *
    * The code contained in this file was originally taken from QtPdfWidgets.
    * Various changes have been made. All bugs are to be reported to CorePDF.
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, vsit http://www.gnu.org/licenses/.
    *
*/

#ifndef QPDFVIEW_P_H
#define QPDFVIEW_P_H

#include "qpdfview.h"

#include <QPointer>
#include <QThread>
#include <QtWidgets/private/qabstractscrollarea_p.h>

QT_BEGIN_NAMESPACE;

class QPdfPageRenderer;
class QPdfDocumentRenderOptions;
class QPdfSearchModel;

class PdfSearchThread : public QThread
{
    Q_OBJECT;

public:
    PdfSearchThread( QObject *parent = nullptr );
    ~PdfSearchThread();

    /** This will reset the search */
    void setDocument( QPdfDocument *doc );

    /** Set the search string. Call searchPage(...) after this */
    void setSearchString( QString );

    /** Start or queue the search */
    void searchPage( int pageNo );

    /** Call searchPage(...) before calling results(...) */
    QVector<QRectF> results( int );

private:
    QPdfDocument *mDoc;
    QPdfSearchModel *searcher;
    QString needle;

    QVector<int> pages;

    QHash<int, QVector<QRectF>> mResults;

    /** Stop completely */
    bool stop;

    /** Stop-other-loops-and-do-this_page flag */
    bool stop_others;

    /** Matches found so far */
    int matchCount;

protected:
    /** We perform the actual search here, and emit the signal */
    void run();

Q_SIGNALS:
    /** Results of @pageNo are ready */
    void resultsReady( int pageNo );

    void matchesFound( int );
    void searchComplete( int numMatches );

    /** Internal signal - used for prioritizing a particular page */
    void pendingRestart();
};

class QPdfViewPrivate : public QAbstractScrollAreaPrivate
{
    Q_DECLARE_PUBLIC(QPdfView);

public:
    QPdfViewPrivate();
    void init();

    void documentStatusChanged();
    void currentPageChanged(int currentPage);
    void calculateViewport();
    void setViewport(QRect viewport);
    void updateScrollBars();

    void pageRendered(int pageNumber, QSize imageSize, const QImage &image, quint64 requestId);
    void invalidateDocumentLayout();
    void invalidatePageCache();

    qreal yPositionForPage(int page) const;

    qreal zoomFactor() const;
    void searchText( QString text, int pageNo );

    struct DocumentLayout
    {
        QSize documentSize;
        QHash<int, QRect> pageGeometries;
    };

    DocumentLayout calculateDocumentLayout() const;
    DocumentLayout calculateDocumentLayoutSingle() const;
    DocumentLayout calculateDocumentLayoutFacing() const;
    DocumentLayout calculateDocumentLayoutBook() const;
    DocumentLayout calculateDocumentLayoutOverview() const;

    void updateDocumentLayout();

    void paintSearchRects( QPainter *, int );

    QPointer<QPdfDocument> m_document;
    QPdfPageNavigation* m_pageNavigation;

    QPdfPageRenderer *m_pageRenderer;
    QPdfDocumentRenderOptions m_renderOpts;

    bool m_continuous;
    QPdfView::PageLayout m_pageLayout;
    QPdfView::ZoomMode m_zoomMode;
    qreal m_zoomFactor;

    int m_pageSpacing;
    QMargins m_documentMargins;

    bool m_blockPageScrolling;

    QMetaObject::Connection m_documentStatusChangedConnection;

    QRect m_viewport;

    QHash<int, QImage> m_pageCache;
    QVector<int> m_cachedPagesLRU;
    int m_pageCacheLimit;

    DocumentLayout m_documentLayout;

    qreal m_screenResolution; // pixels per point

    PdfSearchThread *m_searchThread;
    QHash<int, QVector<QRectF>> searchRects;
};

Q_DECLARE_TYPEINFO(QPdfViewPrivate::DocumentLayout, Q_MOVABLE_TYPE);

QT_END_NAMESPACE;

#endif // QPDFVIEW_P_H
